function InstanceError(name) {
  return "Inherited interface "+name+" is not instantiated";
}

function Repository() {}
Repository.prototype.adopt = function(object) {
  for(var name in object) if(object.hasOwnProperty(name))
    this[name] = object[name];
};

function Module(/* ... */) {
  this.componentInfos = arguments;
  var componentInfosByCID = {};
  for(var index = 0; index < this.componentInfos.length; index++) {
    var componentInfo = this.componentInfos[index];
    componentInfosByCID[componentInfo.cid] = componentInfo;
  }
  this.componentInfosByCID = componentInfosByCID;
}
Module.prototype.registerSelf = function(compMgr, fileSpec, location, type) {
  compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
  for(var index = 0; index < this.componentInfos.length; index++) {
    var componentInfo = this.componentInfos[index];
    compMgr.registerFactoryLocation(componentInfo.cid, componentInfo.name, componentInfo.contractID, fileSpec, location, type);
  }
};
Module.prototype.unregisterSelf = function(compMgr, fileSpec, location) {
  compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
  for(var index = 0; index < this.componentInfos.length; index++) {
    var componentInfo = this.componentInfos[index];
    compMgr.unregisterFactoryLocation(componentInfo.cid, fileSpec);
  }
};
Module.prototype.getClassObject = function(compMgr, cid, iid) {
  if(!iid.equals(Components.interfaces.nsIFactory))
    throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
  if(!this.componentInfosByCID.hasOwnProperty(cid))
    throw Components.results.NS_ERROR_NO_INTERFACE;
  return new Factory(this.componentInfosByCID[cid]);
};
Module.prototype.canUnload = function(compMgr) {
  return true;
};

function ComponentInfo(constructor, name, contractID, cid, iids) {
  if(!(constructor.prototype instanceof Component))
    throw "constructor must inherit Component";
  this.constructor = constructor;
  this.name = name;
  this.contractID = contractID;
  if(typeof cid == "string") cid = Components.ID(cid);
  this.cid = cid;
  this.iids = constructor.prototype.iids;
  this.iidMap = constructor.prototype.iidMap;
}

function Component(/* ... */) {
  var iids = [];
  var iidMap = {};
  for(var index = 0; index < arguments.length; index++) {
    var iid = arguments[index];
    if(typeof iid == "string") iid = Components.interfaces[iid];
    iids[iids.length] = iid;
    iidMap[iid] = iid;
  }
  this.iids = iids;
  this.iidMap = iidMap;
}
Component.prototype.hasInterface = function(iid) {
  if(iid.equals(Components.interfaces.nsISupports))
    return true;
  if(this.iidMap.hasOwnProperty(iid))
    return true;
  for(var iids = this.iids, index = 0; index < iids.length; index++)
    if(iid.equals(iids[index]))
      return true;
  return false;
};
Component.prototype.QueryInterface = function(iid) {
  if(!this.hasInterface(iid)) throw Components.results.NS_ERROR_NO_INTERFACE;
  return this;
};
Component.prototype.extend = function(/* ... */) {
  for(var index = arguments.length-1; index >= 0; index--) {
    var iid = this.iidMap[arguments[index]],
      componentInfo = Components.classes["@mozilla.org/scriptableInterfaceInfo;1"].
        createInstance(Components.interfaces.nsIScriptableInterfaceInfo);
    try { componentInfo.init(iid); }
    catch(e) { throw "Invalid interface"; }
    var name = componentInfo.name;

    var methodNameMap = {};

    for(var methodIndex = 0, methodCount = componentInfo.methodCount; methodIndex < methodCount; methodIndex++) {
      var methodInfo = componentInfo.getMethodInfo(methodIndex),
        methodName = methodInfo.name;

      if(this.hasOwnProperty(methodName) && !methodNameMap.hasOwnProperty(methodName)) continue;
      if(methodInfo.isHidden || methodInfo.isNotXPCOM || methodInfo.isConstructor) continue;

      methodNameMap[methodName] = true;

      if(methodInfo.isGetter) {
        var callee = function() {
          var callee = arguments.callee;
          var instance = this[callee.interfaceName];
          if(instance == null) throw new InstanceError(callee.interfaceName);
          return instance[callee.methodName];
        }
        callee.methodName = methodName;
        callee.interfaceName = name;
        this[methodName] getter = callee;
      }
      else if(methodInfo.isSetter) {
        var callee = function(value) {
          var callee = arguments.callee;
          var instance = this[callee.interfaceName];
          if(instance == null) throw new InstanceError(callee.interfaceName);
          instance[callee.methodName] = value;
          return true;
        }
        callee.methodName = methodName;
        callee.interfaceName = name;
        this[methodName] setter = callee;
      }
      else {
        var callee = function() {
          var callee = arguments.callee;
          var instance = this[callee.interfaceName];
          if(instance == null) throw new InstanceError(callee.interfaceName);
          return instance[callee.methodName].apply(instance, arguments);
        }
        callee.methodName = methodName;
        callee.interfaceName = name;
        this[methodName] = callee;
      }
    }
  }

  this[name] = null;
};

function ComponentExtension(constructor/* , ... */) {
  var component = new constructor;
  if(!(component instanceof Component)) throw "Invalid component constructor";

  var iids = component.iids;
  var iidMap = component.iidMap;
  for(var index = 1; index < arguments.length; index++) {
    var iid = arguments[index];
    if(typeof iid == "string") iid = Components.interfaces[iid];
    if(component.hasInterface(iid)) continue;
    iids[iids.length] = iid;
    iidMap[iid] = iid;
  }
  return component;
}
ComponentExtension.prototype = new Component;

function Implementor(/* ... */) {
  Component.apply(this, arguments);
  Component.prototype.extend.apply(this, arguments);
}
Implementor.prototype = new Component();

function Factory(componentInfo) {
  this.componentInfo = componentInfo;
}
Factory.prototype.createInstance = function(outer, iid) {
  var componentInfo = this.componentInfo;
  if(outer != null)
    throw Components.results.NS_ERROR_NO_AGGREGATION;
  // XXX make this work like in QueryInterface:
  if((!componentInfo.iidMap.hasOwnProperty(iid) || !iid.equals(componentInfo.iidMap[iid])) && !iid.equals(Components.interfaces.nsISupports))
    ;//throw Components.results.NS_ERROR_NO_INTERFACE;
  var instance = new componentInfo.constructor();
  instance.componentInfo = this.componentInfo;
  return instance;
};

function toString() {
  return "[class.js]";
}

// XXX remove
var windowManager = Components.classes["@mozilla.org/appshell/window-mediator;1"].
  getService(Components.interfaces.nsIWindowMediator);
var window = windowManager.getMostRecentWindow("mail:3pane");

var EXPORTED_SYMBOLS = ["ComponentInfo", "Component", "ComponentExtension", "Implementor", "Module", "Factory", "Repository", "window"];
